
#Import "toybox/toybox"

Using std..
Using mojo..
Using mojo3d..

Using ToyBox..

Const SHADOWS_ON:Bool		= True

Const TOWER_WIDTH:Int		= 4
Const TOWER_HEIGHT:Int		= 10

Class Canvas Extension

	Method ShadowText (text:String, tx:Float, ty:Float, color:Color = Null, handleX:Float = 0.0, handleY:Float = 0.0)
		If Not color Then color = Color.White
		Color = Color.Black
		DrawText (text, tx + 1, ty + 1, handleX, handleY)
		Color = color
		DrawText (text, tx, ty, handleX, handleY)
	End
	
End

Class Game Extends Window

	Const CAMERA_SPEED_NORMAL:Float	= 0.1
	Const CAMERA_SPEED_TURBO:Float	= 1.0

	Field boost_camera:Bool	' Camera speed boost (left-shift sets True)
	
	Field scene:Scene
	Field camera:Camera
	Field light:Light
	
	Field center:Model		' World centre/center! Gets shifted up in demo, just for directing light and camera...

	' Wall-specific demo stuff...

	Field wall_left:BrickWall
	Field wall_right:BrickWall
	Field wall_front:BrickWall
	Field wall_back:BrickWall
	
	Method New (title:String, width:Int, height:Int, flags:WindowFlags)
		
		Super.New (title, width, height, flags)
		
		scene = Scene.GetCurrent ()
		
		center = New Model
		
			center.Move (0, 15, 0)
		
		camera = New Camera

			camera.Near	= 0.1
			camera.Far	= 500
			
			camera.Move (0, 30, -20)
			camera.PointAt (center)
			
		light = New Light
			
			light.Type = LightType.Directional
			light.CastsShadow = SHADOWS_ON
			
			light.Move (-100, 100, 100)
			light.PointAt (center)

		Local ground:GroundBody = New GroundBody (, , , New PbrMaterial (Color.Green * 0.3))
		
			ground.Start ()

		BuildBrickShotHouse (TOWER_WIDTH, TOWER_HEIGHT, , , , 1.0, 0.0) ' Last two params: mass and brick 'jiggle'
		
	End

	Method ProcessInput:Void ()

		If Keyboard.KeyHit (Key.F5)
		
			For Local b:BoxBody = Eachin ToyBox.BoxList
				b.model.Destroy ()
				ToyBox.BoxList.Remove (b)
			Next
			
			For Local b:SphereBody = Eachin ToyBox.SphereList
				b.model.Destroy ()
				ToyBox.SphereList.Remove (b)
			Next
			
			For Local b:CapsuleBody = Eachin ToyBox.CapsuleList
				b.model.Destroy ()
				ToyBox.CapsuleList.Remove (b)
			Next
			
			For Local b:ConeBody = Eachin ToyBox.ConeList
				b.model.Destroy ()
				ToyBox.ConeList.Remove (b)
			Next
			
			For Local b:CylinderBody = Eachin ToyBox.CylinderList
				b.model.Destroy ()
				ToyBox.CylinderList.Remove (b)
			Next
			
			BuildBrickShotHouse (TOWER_WIDTH, TOWER_HEIGHT * 3, , , , 1.0, 3.0)

			light.CastsShadow = False

		Endif

		If Keyboard.KeyHit (Key.Escape) Then App.Terminate ()
		
		' Shoot from camera position with force, size, mass:
		
		If Keyboard.KeyHit (Key.Space) Then Shoot (camera, 100.0, 2.0, 1.0)
		' Function Shoot (shoot_from:Entity = Null, force:Float = 100.0, size:Float = 2.0, mass:Float = 1.0, light:Light)
		
		If Keyboard.KeyHit (Key.S) Then light.CastsShadow = Not light.CastsShadow

		If Keyboard.KeyDown (Key.LeftShift)
			boost_camera = True
		Else
			boost_camera = False
		Endif
		
		If Keyboard.KeyDown (Key.A)
			If boost_camera
				camera.Move (0.0, 0.0, CAMERA_SPEED_TURBO)
			Else
				camera.Move (0.0, 0.0, CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Z)
			If boost_camera
				camera.Move (0.0, 0.0, -CAMERA_SPEED_TURBO)
			Else
				camera.Move (0.0, 0.0, -CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Left)
			camera.Rotate (0.0, 1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Right)
			camera.Rotate (0.0, -1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Up)
			camera.Rotate (1.0, 0.0, 0.0, True)
		Endif
 
		If Keyboard.KeyDown (Key.Down)
			camera.Rotate (-1.0, 0.0, 0.0, True)
		Endif
		
		camera.FOV = camera.FOV - (Mouse.WheelY * 5.0)
		
	End
	
	' Wall width/height = each face...
	
	' 'width' should be even number or looks weird...
	
	Method BuildBrickShotHouse (wall_width:Float = 6, wall_height:Float = 6, brick_width:Float = 5, brick_height:Float = 2.5, brick_depth:Float = 2.5, brick_mass:Float = 1.0, jiggle:Float = 0.0)

			Local size:Float = wall_width * brick_width
			Local half:Float = size * 0.5
			
			Local half_brick_width:Float = brick_width * 0.5
			
			' Note: 1.0, 2.0 here are brick mass and random 'jiggle' (yaw to twist each brick slightly for
			' more realistic look). More jiggle means less stable initially!
			
			wall_front = New BrickWall (wall_width, wall_height, brick_width, brick_height, brick_depth, brick_mass, jiggle)
		
				wall_front.Start ()
		
			wall_back = New BrickWall (wall_width, wall_height, brick_width, brick_height, brick_depth, brick_mass, jiggle)
		
				wall_back.Move (-brick_width, 0, size)
				wall_back.Rotate (0, 180, 0)
				wall_back.Start ()
		
			wall_left = New BrickWall (wall_width, wall_height, brick_width, brick_height, brick_depth, brick_mass, jiggle)
		
				wall_left.Move (-(half + half_brick_width), 0, half - half_brick_width)
				wall_left.Rotate (0, -90, 0)
				wall_left.Start ()
		
			wall_right= New BrickWall (wall_width, wall_height, brick_width, brick_height, brick_depth, brick_mass, jiggle)
		
				wall_right.Move (half - half_brick_width, 0, half + half_brick_width)
				wall_right.Rotate (0, 90, 0)
				wall_right.Start ()

	End
		
	Method UpdateGame:Void ()

		For Local brick:BoxBody = Eachin ToyBox.BoxList
			brick.SetStateColor ()
		Next
		
		ProcessInput ()
		scene.Update ()

	End
	
	Method OnRender (canvas:Canvas) Override
 
 		UpdateGame ()
		RequestRender ()

		scene.Render (canvas)

		' ShadowText is in class extension at top!
		
		canvas.ShadowText ("FPS: " + App.FPS,									20.0, 20.0)
		
		canvas.ShadowText ("A/Z + cursors to move; left-shift to boost",		20.0, 50.0)
		canvas.ShadowText ("Space to shoot!",									20.0, 70.0)
		canvas.ShadowText ("F5 to reset scene",									20.0, 90.0)
		canvas.ShadowText ("S to toggle shadows",								20.0, 110.0)
		
		canvas.ShadowText ("Boxes: "			+ ToyBox.BoxList.Count (),		20.0, 150.0)

		canvas.ShadowText ("KEY:",												Window.Width - 200, 20)
		canvas.ShadowText ("Active, but nodding off...",						Window.Width - 200, 40,		ToyBox.PHYS_NODDING_OFF)
		canvas.ShadowText ("Sleeping",											Window.Width - 200, 60,		ToyBox.PHYS_SLEEPING)
		canvas.ShadowText ("Never sleeps",										Window.Width - 200, 80,		ToyBox.PHYS_NEVER_SLEEPS)
		canvas.ShadowText ("Disabled",											Window.Width - 200, 100,	ToyBox.PHYS_DISABLED)
		canvas.ShadowText ("Other: Active",										Window.Width - 200, 120)
		
		canvas.Color = Color.Black
		canvas.DrawRect (Width * 0.5, Height * 0.5, 2.0, 2.0)

		canvas.Color = Color.White
		canvas.DrawRect (Width * 0.5 - 1.0, Height * 0.5 - 1.0, 2.0, 2.0)

	End

End
 
Function Main ()
 
	Local width:Int			= 640
	Local height:Int		= 480
	Local flags:WindowFlags	= WindowFlags.Center
	
'	Local width:Int			= 1920
'	Local height:Int		= 1080
'	Local flags:WindowFlags	= WindowFlags.Fullscreen
	
	New AppInstance
	
	New Game ("Physics Demo", width, height, flags)
	
	App.Run ()
		
End
