
#Import "toybox/toybox"

Using std..
Using mojo..
Using mojo3d..

Using ToyBox..

Class Canvas Extension

	Method ShadowText (text:String, tx:Float, ty:Float, handleX:Float = 0.0, handleY:Float = 0.0)
		Color = Color.Black
		DrawText (text, tx + 1, ty + 1, handleX, handleY)
		Color = Color.White
		DrawText (text, tx, ty, handleX, handleY)
	End
	
End

Class Game Extends Window

	Const CAMERA_SPEED_NORMAL:Float	= 0.1
	Const CAMERA_SPEED_TURBO:Float	= 1.0

	Field boost_camera:Bool	' Camera speed boost (left-shift sets True)
	
	Field scene:Scene
	Field camera:Camera
	Field light:Light

	Field center:Model		' World centre/center! Gets shifted up in demo, just for directing light and camera...

	' Wall-specific demo stuff...

	Field wall:BrickWall
	
	Field wall_width:Float			= 12
	Field wall_height:Float			= 12
	
	Field wall_brick_width:Float	= 5
	Field wall_brick_height:Float	= 2.5
	Field wall_brick_depth:Float	= 2.5
	
	Method New (title:String, width:Int, height:Int, flags:WindowFlags)
		
		Super.New (title, width, height, flags)
		
		scene = Scene.GetCurrent ()
		
		center = New Model
		
			center.Move (0, 15, 0)
		
		camera = New Camera

			camera.Near	= 0.1
			camera.Far	= 500
			
			camera.Move (0, 5, -20)
			camera.PointAt (center)
			
		light = New Light
			
			light.Type = LightType.Directional
			light.CastsShadow = True
			
			light.Move (-100, 100, -100)
			light.PointAt (center)

		Local ground:GroundBody = New GroundBody (, , , New PbrMaterial (Color.Green * 0.3))
		
			ground.Start ()

		Local size:Float = wall_width * wall_brick_width
		Local half:Float = size * 0.5
		
		Local half_brick_width:Float = wall_brick_width * 0.5
		
		Local wall:BrickWall = New BrickWall (wall_width, wall_height, wall_brick_width, wall_brick_height, wall_brick_depth, 1.0, 2.0)
	
			wall.Start ()
		
	End
	
	Method ProcessInput:Void ()

		If Keyboard.KeyHit (Key.F5)
		
			For Local b:BoxBody = Eachin ToyBox.BoxList
				b.model.Destroy ()
				ToyBox.BoxList.Remove (b)
			Next
			
			For Local b:SphereBody = Eachin ToyBox.SphereList
				b.model.Destroy ()
				ToyBox.SphereList.Remove (b)
			Next
			
			For Local b:CapsuleBody = Eachin ToyBox.CapsuleList
				b.model.Destroy ()
				ToyBox.CapsuleList.Remove (b)
			Next
			
			For Local b:ConeBody = Eachin ToyBox.ConeList
				b.model.Destroy ()
				ToyBox.ConeList.Remove (b)
			Next
			
			For Local b:CylinderBody = Eachin ToyBox.CylinderList
				b.model.Destroy ()
				ToyBox.CylinderList.Remove (b)
			Next
			
			wall = New BrickWall (wall_width, wall_height, wall_brick_width, wall_brick_height, wall_brick_depth, 1.0, 2.0)
		
				wall.Move (0, 0, 5)
			
				wall.Start ()
			
		Endif

		If Keyboard.KeyHit (Key.Escape) Then App.Terminate ()
		
		' Shoot from camera position with force, size, mass:
		
		If Keyboard.KeyHit (Key.Space) Then Shoot (camera, 100.0, 2.0, 1.0)
		
		If Keyboard.KeyHit (Key.S) Then light.CastsShadow = Not light.CastsShadow

		If Keyboard.KeyDown (Key.LeftShift)
			boost_camera = True
		Else
			boost_camera = False
		Endif
		
		If Keyboard.KeyDown (Key.A)
			If boost_camera
				camera.Move (0.0, 0.0, CAMERA_SPEED_TURBO)
			Else
				camera.Move (0.0, 0.0, CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Z)
			If boost_camera
				camera.Move (0.0, 0.0, -CAMERA_SPEED_TURBO)
			Else
				camera.Move (0.0, 0.0, -CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Left)
			camera.Rotate (0.0, 1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Right)
			camera.Rotate (0.0, -1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Up)
			camera.Rotate (1.0, 0.0, 0.0, True)
		Endif
 
		If Keyboard.KeyDown (Key.Down)
			camera.Rotate (-1.0, 0.0, 0.0, true)
		Endif
		
		camera.FOV = camera.FOV - (Mouse.WheelY * 5.0)
		
	End
	
	Method UpdateGame:Void ()

		ProcessInput ()
		scene.Update ()

	End
	
	Method OnRender (canvas:Canvas) Override
 
 		UpdateGame ()
		RequestRender ()
		scene.Render (canvas)

		' ShadowText is in class extension at top!
		
		canvas.ShadowText ("FPS: " + App.FPS,									20.0, 20.0)
		
		canvas.ShadowText ("A/Z + cursors to move; left-shift to boost",		20.0, 50.0)
		canvas.ShadowText ("Space to shoot!",									20.0, 70.0)
		canvas.ShadowText ("F5 to reset scene",									20.0, 90.0)
		canvas.ShadowText ("S to toggle shadows",								20.0, 110.0)
		
		canvas.ShadowText ("Boxes: "			+ ToyBox.BoxList.Count (),		20.0, 150.0)

		canvas.Color = Color.Black
		canvas.DrawRect (Width * 0.5, Height * 0.5, 2.0, 2.0)

		canvas.Color = Color.White
		canvas.DrawRect (Width * 0.5 - 1.0, Height * 0.5 - 1.0, 2.0, 2.0)
		
'		For Local b:BoxBody = Eachin ToyBox.BoxList
'			If CheckVisible (b.model, camera)
'				Local pos:Vec2f = camera.ProjectToViewport (b.model.Position)
'				canvas.DrawText ("X", pos.x, Height - pos.y)
'			Endif
'		Next
'		
'		For Local b:SphereBody = Eachin ToyBox.SphereList
'			If CheckVisible (b.model, camera)
'				Local pos:Vec2f = camera.ProjectToViewport (b.model.Position)
'				canvas.DrawText ("X", pos.x, Height - pos.y)
'			Endif
'		Next
'		
'		For Local b:CylinderBody = Eachin ToyBox.CylinderList
'			If CheckVisible (b.model, camera)
'				Local pos:Vec2f = camera.ProjectToViewport (b.model.Position)
'				canvas.DrawText ("X", pos.x, Height - pos.y)
'			Endif
'		Next
'		
'		For Local b:ConeBody = Eachin ToyBox.ConeList
'			If CheckVisible (b.model, camera)
'				Local pos:Vec2f = camera.ProjectToViewport (b.model.Position)
'				canvas.DrawText ("X", pos.x, Height - pos.y)
'			Endif
'		Next
'		
'		For Local b:CapsuleBody = Eachin ToyBox.CapsuleList
'			If CheckVisible (b.model, camera)
'				Local pos:Vec2f = camera.ProjectToViewport (b.model.Position)
'				canvas.DrawText ("X", pos.x, Height - pos.y)
'			Endif
'		Next
'		
	End

End
 
Function Main ()
 
	Local width:Int			= 640
	Local height:Int		= 480
	Local flags:WindowFlags	= WindowFlags.Center
	
'	Local width:Int			= 1920
'	Local height:Int		= 1080
'	Local flags:WindowFlags	= WindowFlags.Fullscreen
	
	New AppInstance
	
	New Game ("Physics Demo", width, height, flags)
	
	App.Run ()
		
End
