#Import "toybox/toybox"




	' NOTE:

	' Dropped boxes are BoxBody type (BoxCollider)

	' *** All other bodies are MeshColliders! ***

	' F5 to re-drop.


Using std..
Using mojo..
Using mojo3d..

Using ToyBox..

Const SCENE_UPDATE:Float = 60.0
 
Class Game Extends Window

	Field scene:Scene
	Field camera:Camera
	Field light:Light

	Field pivot:Model
	
	Field dropped:BoxBody
	
	Method New (title:String, width:Int, height:Int, flags:WindowFlags)
		
		Super.New (title, width, height, flags)

		SeedRnd (Millisecs ())
				
		scene = Scene.GetCurrent ()
		
			scene.UpdateRate = SCENE_UPDATE ' Still nope!
			
		pivot = New Model
		
		camera = New Camera

			camera.Near	= 0.1
			camera.Far	= 500
			
			camera.Move (0, 5, -10)
			
		light = New Light
			
			light.Type = LightType.Directional
			light.CastsShadow = True
			
			light.Move (-100, 100, -100)

		Local ground:GroundBody = New GroundBody
		
			ground.Start ()

		DropBoxes ()
		
		Local size:Float = 3.0
		
		Local box_body:MeshBody
		Local sphere_body:MeshBody
		Local cone_body:MeshBody
		
		Local box_model:Model		= Model.CreateBox (New Boxf (-size * 0.5, - size * 0.5, - size * 0.5, size * 0.5, size * 0.5, size * 0.5), 1, 1, 1, New PbrMaterial (Color.Red))
		Local sphere_model:Model	= Model.CreateSphere (size * 0.5, 32, 32, New PbrMaterial (Color.Red))
		Local cone_model:Model		= Model.CreateCone (size * 0.5, 4.0, Axis.Y, 256, New PbrMaterial (Color.Red))
		
		Local x:Float
		
		box_body					= New MeshBody (box_model)
		
			x = -size * 2
			
			box_body.Move		(x, size * 0.5, 0)
			box_body.Rotate		(45, 0, 0)
			box_body.SetBounce	(0.0)
			
			box_body.Start ()

		sphere_body					= New MeshBody (sphere_model)
			
			x = 0
			
			sphere_body.Move		(x, size * 0.5, 0)
			sphere_body.Rotate		(0, 0, 0)
			sphere_body.SetBounce	(0.0)
			
			sphere_body.Start ()

		cone_body					= New MeshBody (cone_model)
			
			x = size * 2
			
			cone_body.Move			(x, size, 0)
			cone_body.Rotate		(180, 0, 0)
			cone_body.SetBounce		(0.0)
			
			cone_body.Start ()

			' TODO: Calling sphere_body.Start again in error freezes program!
			
		camera.PointAt (pivot)
		light.PointAt (pivot)

	End

	Method DropBoxes ()

		For Local loop:Int = 1 To 500
		
			' Dropped body...
			
			' Parameters below are width, height, depth and mass, plus material. Experiment with them!
			
			Local dropped_size:Float = Rnd (0.1, 1)
			
			dropped = New BoxBody (dropped_size, dropped_size, dropped_size, 1, New PbrMaterial (Color.White))
				
				' IMPORTANT: Set up position, etc, BEFORE starting physics. Play with these!
				
				dropped.Move		(Rnd (-10, 10), Rnd (5, 25), 0)
				dropped.Rotate		(Rnd (360.0), Rnd (360.0), Rnd (360.0))
				dropped.SetBounce	(0.1)
				
				' Start body physics:
				
				dropped.Start ()

		Next
	
	End

	Method ProcessInput:Void ()

		If Keyboard.KeyHit (Key.Escape) Then App.Terminate ()
		
		If Keyboard.KeyHit (Key.F5)
			
			For Local b:BoxBody = Eachin ToyBox.BoxList
				' TODO: Add single command!
				b.model.Destroy ()
				ToyBox.BoxList.Remove (b)
			Next
			
			DropBoxes ()
			
		Endif
		
	End
	
	Method UpdateGame:Void ()

		Title = "MeshColliders; updating scene at " + scene.UpdateRate + " FPS"

		ProcessInput ()
		scene.Update ()

	End
	
	Method OnRender (canvas:Canvas) Override
 
 		UpdateGame ()
		RequestRender ()
		scene.Render (canvas)
		
	End

End
 
Function Main ()
 
	Local width:Int			= 640
	Local height:Int		= 480
	Local flags:WindowFlags	= WindowFlags.Center
	
'	Local width:Int			= 1920
'	Local height:Int		= 1080
'	Local flags:WindowFlags	= WindowFlags.Fullscreen
	
	New AppInstance
	
	New Game ("MeshColliders: ", width, height, flags)
	
	App.Run ()
		
End
