
#Import "toybox/toybox"

Using std..
Using mojo..
Using mojo3d..

Using ToyBox..

Const SHADOWS_ON:Bool		= True

Class Canvas Extension

	Method ShadowText (text:String, tx:Float, ty:Float, color:Color = Null, handleX:Float = 0.0, handleY:Float = 0.0)
		If Not color Then color = Color.White
		Color = Color.Black
		DrawText (text, tx + 1, ty + 1, handleX, handleY)
		Color = color
		DrawText (text, tx, ty, handleX, handleY)
	End
	
End

Class Game Extends Window

	Const CAMERA_SPEED_NORMAL:Float	= 0.1
	Const CAMERA_SPEED_TURBO:Float	= 1.0

	Field boost_camera:Bool	' Camera speed boost (left-shift sets True)
	
	Field scene:Scene
	Field camera:Camera
	Field light:Light

	Field center:Model		' World centre/center! Gets shifted up in demo, just for directing light and camera...

	Field rocket:ConeBody
	
	Method New (title:String, width:Int, height:Int, flags:WindowFlags)
		
		Super.New (title, width, height, flags)
		
		scene = Scene.GetCurrent ()
		
		center = New Model
		
			center.Move (0, 5, 0)
		
		camera = New Camera

			camera.Near	= 0.1
			camera.Far	= 500
			
			camera.Move (0, 10, -15)
			camera.PointAt (center)
			
		light = New Light
			
			light.Type = LightType.Directional
			light.CastsShadow = SHADOWS_ON
			
			light.Move (-100, 100, 100)
			light.PointAt (center)

		Local ground:GroundBody = New GroundBody (, , , New PbrMaterial (Color.Green * 0.3))
		
			ground.Start ()

		rocket = New ConeBody (2, 8, 1, New PbrMaterial (Color.Silver))
		
			rocket.Start ()
		
	End

	Method ProcessInput:Void ()

		If Keyboard.KeyHit (Key.F5)
		
			For Local b:BoxBody = Eachin ToyBox.BoxList
				b.model.Destroy ()
				ToyBox.BoxList.Remove (b)
			Next
			
			For Local b:SphereBody = Eachin ToyBox.SphereList
				b.model.Destroy ()
				ToyBox.SphereList.Remove (b)
			Next
			
			For Local b:CapsuleBody = Eachin ToyBox.CapsuleList
				b.model.Destroy ()
				ToyBox.CapsuleList.Remove (b)
			Next
			
			For Local b:ConeBody = Eachin ToyBox.ConeList
				b.model.Destroy ()
				ToyBox.ConeList.Remove (b)
			Next
			
			For Local b:CylinderBody = Eachin ToyBox.CylinderList
				b.model.Destroy ()
				ToyBox.CylinderList.Remove (b)
			Next

			rocket = New ConeBody (2, 8, 1, New PbrMaterial (Color.Silver))
		
				rocket.Start ()

		Endif

		If Keyboard.KeyHit (Key.Escape) Then App.Terminate ()
		
		' Shoot from camera position with force, size, mass:
		
		If Keyboard.KeyDown (Key.Space) Then rocket.Boost (0.0, 0.25, 0.0)
		
		If Keyboard.KeyHit (Key.S) Then light.CastsShadow = Not light.CastsShadow

		If Keyboard.KeyDown (Key.LeftShift)
			boost_camera = True
		Else
			boost_camera = False
		Endif
		
		If Keyboard.KeyDown (Key.A)
			If boost_camera
				camera.Move (0.0, 0.0, CAMERA_SPEED_TURBO)
			Else
				camera.Move (0.0, 0.0, CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Z)
			If boost_camera
				camera.Move (0.0, 0.0, -CAMERA_SPEED_TURBO)
			Else
				camera.Move (0.0, 0.0, -CAMERA_SPEED_NORMAL)
			Endif
		Endif
 
		If Keyboard.KeyDown (Key.Left)
			camera.Rotate (0.0, 1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Right)
			camera.Rotate (0.0, -1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Up)
			camera.Rotate (1.0, 0.0, 0.0, True)
		Endif
 
		If Keyboard.KeyDown (Key.Down)
			camera.Rotate (-1.0, 0.0, 0.0, True)
		Endif
		
		camera.FOV = camera.FOV - (Mouse.WheelY * 5.0)
		
	End

	Method UpdateGame:Void ()

'		For Local brick:BoxBody = Eachin ToyBox.BoxList
'			brick.SetStateColor ()
'		Next
		
		camera.PointAt (rocket.model)
		
		ProcessInput ()
		scene.Update ()

	End
	
	Method OnRender (canvas:Canvas) Override
 
 		UpdateGame ()
		RequestRender ()

		scene.Render (canvas)

		' ShadowText is in class extension at top!
		
		canvas.ShadowText ("FPS: " + App.FPS,									20.0, 20.0)
		
		canvas.ShadowText ("A/Z + cursors to move; left-shift to boost",		20.0, 50.0)
		canvas.ShadowText ("Space to boost rocket",								20.0, 70.0)
		canvas.ShadowText ("F5 to reset scene",									20.0, 90.0)
		canvas.ShadowText ("S to toggle shadows",								20.0, 110.0)
		
'		canvas.ShadowText ("Boxes: "			+ ToyBox.BoxList.Count (),		20.0, 150.0)

'		canvas.ShadowText ("KEY:",												Window.Width - 200, 20)
'		canvas.ShadowText ("Active, but nodding off...",						Window.Width - 200, 40,		ToyBox.PHYS_NODDING_OFF)
'		canvas.ShadowText ("Sleeping",											Window.Width - 200, 60,		ToyBox.PHYS_SLEEPING)
'		canvas.ShadowText ("Never sleeps",										Window.Width - 200, 80,		ToyBox.PHYS_NEVER_SLEEPS)
'		canvas.ShadowText ("Disabled",											Window.Width - 200, 100,	ToyBox.PHYS_DISABLED)
'		canvas.ShadowText ("Other: Active",										Window.Width - 200, 120)
		
	End

End
 
Function Main ()
 
	Local width:Int			= 640
	Local height:Int		= 480
	Local flags:WindowFlags	= WindowFlags.Center
	
'	Local width:Int			= 1920
'	Local height:Int		= 1080
'	Local flags:WindowFlags	= WindowFlags.Fullscreen
	
	New AppInstance
	
	New Game ("Physics Demo", width, height, flags)
	
	App.Run ()
		
End
