
#Import "toybox/toybox"

Using std..
Using mojo..
Using mojo3d..

Using ToyBox..
 
Class Game Extends Window

	Field scene:Scene
	Field camera:Camera
	Field light:Light

	Field box:BoxBody

	Method New (title:String, width:Int, height:Int, flags:WindowFlags)
		
		Super.New (title, width, height, flags)
		
		scene = Scene.GetCurrent ()
		
		camera = New Camera

			camera.Near	= 0.1
			camera.Far	= 500
			
			camera.Move (0, 5, -10)
			
		light = New Light
			
			light.Type = LightType.Directional
			light.CastsShadow = True
			
			light.Move (-100, 100, -100)

		Local ground:GroundBody = New GroundBody
		
			ground.Start ()

		' Parameters below are width, height, depth and mass, plus material...
		
		box = New BoxBody (1, 1, 1, 1, New PbrMaterial (Color.Red))
			
			' IMPORTANT: Set up body BEFORE starting physics:
			
			box.Move		(0, Rnd (5, 25), 0)
			box.Rotate		(0, 0, 0)
			box.SetBounce	(0.1)
			
			' Start body physics:
			
			box.Start ()

		camera.PointAt (box.model)
		light.PointAt (box.model)

	End
	
	Method ProcessInput:Void ()

		If Keyboard.KeyHit (Key.F5)
		
			' Two-step process to delete at present, Model and RigidBody -- TODO!
			
			box.model.Destroy ()

			ToyBox.BoxList.Remove (box)

			' Now start again! Shake things up with random dimensions and random bounce...
			
			box = New BoxBody (Rnd (1.0, 4.0), Rnd (1.0, 4.0), Rnd (1.0, 4.0), 1.0, New PbrMaterial (Color.Red))
			
				' IMPORTANT: Set up body BEFORE starting physics:
				
				box.Move		(0, Rnd (5, 25), 0)
				box.Rotate		(0, 0, 0)
				box.SetBounce	(Rnd (0.1, 0.95)) ' Higher values will cause 'dancing'!
				
				' Start body physics:
				
				box.Start ()
			
		Endif

		If Keyboard.KeyHit (Key.Escape) Then App.Terminate ()
		
	End
	
	Method UpdateGame:Void ()

		camera.PointAt (box.model)

		ProcessInput ()
		scene.Update ()

	End
	
	Method OnRender (canvas:Canvas) Override
 
 		UpdateGame ()
		RequestRender ()
		scene.Render (canvas)
		
		canvas.DrawText ("FPS: " + App.FPS,	20.0, 20.0)
		canvas.DrawText ("F5 to reset scene",	20.0, 40.0)
		
	End

End
 
Function Main ()
 
	Local width:Int			= 640
	Local height:Int		= 480
	Local flags:WindowFlags	= WindowFlags.Center
	
'	Local width:Int			= 1920
'	Local height:Int		= 1080
'	Local flags:WindowFlags	= WindowFlags.Fullscreen
	
	New AppInstance
	
	New Game ("Simple Demo", width, height, flags)
	
	App.Run ()
		
End
