
#Import "toybox/toybox"
#Import "assets/"

Using std..
Using mojo..
Using mojo3d..

Using ToyBox..

Const DROP:Int = 100

Class Game Extends Window

	Const CAMERA_MOVE:Float = 0.1
	Const SHIFT_BOOST:Float = 4.0
	
	Field cam_boost:Float = 1.0
	 
	Field scene:Scene
	
	Field cam:Camera

	Field light:Light
	Field light_model:Model
	
	Field ground_pixels:Pixmap
	
	Field heightmap:Pixmap
	Field terrain_material:PbrMaterial
	
'	Field terrain_body:MeshBody
	Field terrain_body:TerrainBody
	Field dropped:BoxBody
	
	Method New (title:String, width:Int, height:Int, flags:WindowFlags)

		Super.New (title, width, height, flags)
		
		scene 					= Scene.GetCurrent ()
		scene.AmbientLight		= Color.White * 0.75
		scene.ClearColor		= Color.Sky * 0.75
		
		cam						= New Camera
		cam.Near				= 0.01
		cam.Far					= 1000
		cam.Move (0, 0, -10)
		
		light					= New Light
		light.CastsShadow		= False'True
		light.Range				= 1000
		
		light.Move (0, 50, 10) ' Yes, the sun is 50m above us...
		
		light_model				= Model.CreateSphere (1, 32, 32, New PbrMaterial (Color.Yellow), light)
		light_model.CastsShadow	= False
		
		heightmap = Pixmap.Load ("asset::height.png")
		
		terrain_body = New TerrainBody (heightmap, 512, 32, 512)
		
		' NOTE: Pass own PbrMaterial as next parameter to New TerrainBody () -- defaults
		' to built-in checkerboard material; for example, in place of above line:

'		Local green_material:PbrMaterial = New PbrMaterial (Color.Green)
'		terrain_body = New TerrainBody (heightmap, 512, 32, 512, green_material)
		
			terrain_body.Move (0, -32, 0)
		
			terrain_body.Start ()
		
		light.PointAt (terrain_body.model)

	End

	Method DropBoxes (num:Int = DROP)

		For Local loop:Int = 1 To num
		
			' Dropped body...
			
			' Parameters below are width, height, depth and mass, plus material. Experiment with them!
			
			Local dropped_size:Float = Rnd (0.1, 2.0)
			
			dropped = New BoxBody (dropped_size, dropped_size, dropped_size, 1, ToyBox.RndMaterial)
				
				' IMPORTANT: Set up position, etc, BEFORE starting physics. Play with these!
				
				dropped.Move		(Rnd (-25, 25), Rnd (5, 25), Rnd (25))
				dropped.Rotate		(Rnd (360.0), Rnd (360.0), Rnd (360.0))
				dropped.SetBounce	(0.1)
				
				' Start body physics:
				
				dropped.Start ()

		Next
	
	End

	Method UpdateGame ()

		If Keyboard.KeyHit (Key.S)
			light.CastsShadow = Not light.CastsShadow
		Endif
		
		If Keyboard.KeyHit (Key.F5)

			For Local b:BoxBody = Eachin ToyBox.BoxList
				' TODO: Add single command!
				b.model.Destroy ()
				ToyBox.BoxList.Remove (b)
			Next
			
			DropBoxes ()
			
		Endif
		
		If Keyboard.KeyHit (Key.Escape)
			App.Terminate ()
		Endif

		If Keyboard.KeyDown (Key.LeftShift)
			cam_boost = SHIFT_BOOST
		Else
			cam_boost = 1.0
		Endif
		
		If Keyboard.KeyDown (Key.A)
			cam.Move (0.0, 0.0, CAMERA_MOVE * cam_boost)
		Endif
 
		If Keyboard.KeyDown (Key.Z)
			cam.Move (0.0, 0.0, -CAMERA_MOVE * cam_boost)
		Endif
 
		If Keyboard.KeyDown (Key.Left)
			cam.Rotate (0.0, 1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Right)
			cam.Rotate (0.0, -1.0, 0.0)
		Endif
 
		If Keyboard.KeyDown (Key.Up)
			cam.Rotate (1.0, 0.0, 0.0, True)
		Endif
 
		If Keyboard.KeyDown (Key.Down)
			cam.Rotate (-1.0, 0.0, 0.0, True)
		Endif

	End

	Method ShadowText:Void (canvas:Canvas, s:String, x:Float, y:Float)
		canvas.Color = Color.Black
		canvas.DrawText	(s, x + 1, y + 1)
		canvas.Color = Color.White
		canvas.DrawText	(s, x, y)
	End

	Method RenderText (canvas:Canvas)
		
		ShadowText (canvas, "FPS: " + App.FPS, 20.0, 20.0)
		ShadowText (canvas, "A/Z + Cursors to move camera", 20.0, 60.0)
		ShadowText (canvas, "S to toggle shadows, F5 to drop " + DROP + " boxes", 20.0, 100.0)

	End
	
	Method OnRender (canvas:Canvas) Override

		UpdateGame ()
		
		RequestRender ()
		
		scene.Update ()
		
		scene.Render (canvas)

		RenderText (canvas)
		
	End
	
End

Function Run3D (title:String, width:Int, height:Int, flags:WindowFlags = WindowFlags.Center)

	New AppInstance
	New Game (title, width, height, flags)

	App.Run ()

End

Function Main ()
	Run3D ("3D Scene", 640, 480, WindowFlags.Center)
'	Run3D ("3D Scene", 1920, 1080, WindowFlags.Fullscreen) 
End
