
Namespace ToyBox

#Import "<std>"
#Import "<mojo>"
#Import "<mojo3d>"
#Import "<bullet>"
#Import "<mojo3d-loaders>"
#Import "<assimp>"
#Import "<bullet>"

Const PHYS_SLEEPING:Color		= Color.Orange
Const PHYS_NODDING_OFF:Color	= Color.Yellow
Const PHYS_DISABLED:Color		= Color.Grey
Const PHYS_NEVER_SLEEPS:Color	= Color.Green

Private

' InitParams is used internally to allow creation of box prior to manual placement.

' (Entities should NOT be manually positioned after being attached to a RigidBody.)

' Create body, call body.Move () to set position, then call body.Start ().

' InitParams are only needed for setup, and are freed afterwards, hence not
' added as separate fields.

' Ugh... RoundVolume just a way to get capsule and cone volumes into InitParams!
' TODO: Just re-do as width/height/depth for all volume types?

Class RoundVolume
	Field radius:Float
	Field length:Float
End
		
Class InitParams

	Field mass:Float
	Field group:Short
	Field mask:Short
	Field volume:Variant
	
	' Optional...
	
	Field bounce:Float
	
End

Class BodyBase

	Field model:Model			' mojo3d entity
	Field collider:Collider		' Bullet physics collider
	Field body:RigidBody		' Bullet physics body
	Field init:InitParams		' Setup parameters
	Field live:Bool
	Field init_color:Color

	Method Boost:Void (force_x:Float = 0.0, force_y:Float = 0.0, force_z:Float = 0.0)
		body.ApplyImpulse (model.Basis * New Vec3f (force_x, force_y, force_z))
	End
	
	Method SetStateColor ()
	
		If body And body.btBody

			Select GetState ()
		
				Case bullet.ACTIVE_TAG
					SetColor (init_color)
				Case bullet.ISLAND_SLEEPING
					SetColor (ToyBox.PHYS_SLEEPING)
				Case bullet.WANTS_DEACTIVATION
					SetColor (ToyBox.PHYS_NODDING_OFF)
				Case bullet.DISABLE_DEACTIVATION
					SetColor (ToyBox.PHYS_NEVER_SLEEPS)
				Case bullet.DISABLE_SIMULATION
					SetColor (ToyBox.PHYS_DISABLED)
				
			End
 	
 		Endif
 		
 	End
 	
 	' TODO: Check logic!!!
 	
	Method SetColor:Void (color:Color = Null)
		If model
			If Not color
				init_color = Cast <PbrMaterial> (model.Materials [0]).ColorFactor
			Else
				If Cast <PbrMaterial> (model.Materials [0]).ColorFactor <> color
					model.Materials [0] = New PbrMaterial (color)
				Endif
			Endif
		Endif
	End
	
	Method GetState:Int ()
		Return body.btBody.getActivationState ()
	End
	
	Method GetColor:Color ()
		If model
			Return Cast <PbrMaterial> (model.Materials [0]).ColorFactor
		Endif
		Return New Color (0.0, 0.0, 0.0)
	End
	
	Method Position (x:Float, y:Float, z:Float, local_space:Bool = False)
		model.SetPosition (x, y, z, local_space)
	End
	
	Method Move (x:Float, y:Float, z:Float)
		model.Move (x, y, z)
	End

	Method Rotate (pitch:Float, roll:Float, yaw:Float, localspace:Bool = False)
		model.Rotate (pitch, roll, yaw, localspace)
	End
	
	Method SetBounce (bounce:Float)
	
		If live
			body.Restitution = bounce
		Else
			If init
				init.bounce = bounce
			Endif
		Endif
		
	End
	
End

Public

Function BrickMaterial:PbrMaterial (fixed:Bool = False) ' Test in New BrickWall - no speed difference

	If fixed Return Cast <PbrMaterial> (FixedBrickMaterial)
	
	Local material:PbrMaterial = New PbrMaterial (New Color (Rnd (0.5, 0.75), 0.0, 0.0))

	material.RoughnessFactor = Rnd (0.95, 1.0)

	Return material

End

Global RndMaterial:Material					= New PbrMaterial (Color.None)
Global FixedBrickMaterial:Material			= New PbrMaterial (Color.Red)

Global BoxList:List			<BoxBody>		= New List <BoxBody>
Global SphereList:List		<SphereBody>	= New List <SphereBody>
Global CapsuleList:List		<CapsuleBody>	= New List <CapsuleBody>
Global ConeList:List		<ConeBody>		= New List <ConeBody>
Global CylinderList:List	<CylinderBody>	= New List <CylinderBody>
Global GroundList:List		<GroundBody>	= New List <GroundBody>
Global TerrainList:List		<TerrainBody>	= New List <TerrainBody>

' Global BodyList:List	<BodyBase> []	= New List <BodyBase> [3]
' Add lists to this?

Class SimpleWall

	Field pivot:Model
	
	Field init_list:List <BoxBody>
	
	Method New (width:Int, height:Int, brick_width:Float = 2.0, brick_height:Float = 1.0, brick_depth:Float = 1.0, brick_mass:Float = 1.0, brick_material:Material = ToyBox.RndMaterial)
	
		pivot = New Model
		
		If Not init_list Then init_list = New List <BoxBody>
		
		For Local x:Float = 0 Until width
		
			For Local y:Int = 0 Until height

				Local box:BoxBody = New BoxBody (brick_width, brick_height, brick_depth, brick_mass, brick_material, 1, pivot)
				
				box.SetColor ()
				
				Local x_pos:Float = (x * brick_width) - (width * brick_width * 0.5)
				Local y_pos:Float = (y * brick_height) + (brick_height * 0.5)
				
				box.Move		(x_pos, y_pos, 0.0)
				box.SetBounce	(0.0)
				
				init_list.AddLast (box)
				
			Next
		
		Next
	
	End

	Method Move (x:Float, y:Float, z:Float)
		pivot.Move (x, y, z)
	End
	
	Method Rotate (rx:Float, ry:Float, rz:Float)
		pivot.Rotate (rx, ry, rz)
	End
	
	Method Start ()

		If init_list
			For Local b:BoxBody = Eachin init_list
				b.Start ()
			Next
		Endif

		init_list = Null
		
	End
	
End

Class BrickWall

	Field init_list:List <BoxBody>
	Field pivot:Model
	
	Method New (width:Int, height:Int, brick_width:Float = 2.0, brick_height:Float = 1.0, brick_depth:Float = 1.0, brick_mass:Float = 1.0, jiggle:Float = 0.0)

		pivot = New Model
	
		If Not init_list Then init_list = New List <BoxBody>
		
		For Local x:Float = 0 Until width
		
			For Local y:Int = 0 Until height

				Local box:BoxBody = New BoxBody (brick_width, brick_height, brick_depth, brick_mass, BrickMaterial (), 1, pivot)
				
				box.SetColor ()
				
				Local x_pos:Float = (x * brick_width) - (width * brick_width * 0.5)
				
				If Int (y) & 1
					x_pos = x_pos - (brick_width * 0.25)
				Else
					x_pos = x_pos + (brick_width * 0.25)
				Endif
				
				Local y_pos:Float = (y * brick_height) + (brick_height * 0.5)
				
				box.Move		(x_pos, y_pos, 0.0)
				
				If jiggle
					box.Rotate		(0, Rnd (-jiggle, jiggle), 0)
				Endif
				
				box.SetBounce	(0.0)
				
				init_list.AddLast (box)
				
			Next
		
		Next
	
	End

	Method Move (x:Float, y:Float, z:Float)
		pivot.Move (x, y, z)
	End
	
	Method Rotate (rx:Float, ry:Float, rz:Float)
		pivot.Rotate (rx, ry, rz)
	End
		
	Method Start ()

		If init_list
			For Local b:BoxBody = Eachin init_list
				b.Start ()
			Next
		Endif

		init_list = Null
		
	End
	
End

Class CapsuleBody Extends BodyBase

	Method New (radius:Float = 1.0, length:Float = 1.0, mass:Float = 1.0, material:Material = Null, segments:Int = 32, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= mass
		init.group		= group
		init.mask		= mask

		Local cv:RoundVolume = New RoundVolume
		cv.radius = radius
		cv.length = length
		
		init.volume		= Cast <RoundVolume> (cv)
		
		If Not material
			material = New PbrMaterial (Color.Red)
		Else
			If material = RndMaterial
				material = New PbrMaterial (New Color (Rnd (0.25, 1.0), Rnd (0.25, 1.0), Rnd (0.25, 1.0)), Rnd (1.0), Rnd (1.0))
			Endif
		Endif
		
		SetColor ()
		
		model = Model.CreateCapsule (Cast <RoundVolume> (init.volume).radius, Cast <RoundVolume> (init.volume).length, Axis.Y, segments, material, parent)

		CapsuleList.AddLast (Self)

	End

	Method Start ()

		collider			= New CapsuleCollider (model)
		body				= New RigidBody (model)
 
		Cast <CapsuleCollider> (collider).Radius	= Cast <RoundVolume> (init.volume).radius
		Cast <CapsuleCollider> (collider).Length	= Cast <RoundVolume> (init.volume).length

		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null
		
		live				= True

	End
	
End

Class ConeBody Extends BodyBase

	Method New (radius:Float = 1.0, length:Float = 1.0, mass:Float = 1.0, material:Material = Null, segments:Int = 16, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= mass
		init.group		= group
		init.mask		= mask

		Local cv:RoundVolume = New RoundVolume
		cv.radius = radius
		cv.length = length
		
		init.volume		= Cast <RoundVolume> (cv)
		
		If Not material
			material = New PbrMaterial (Color.Red)
		Else
			If material = RndMaterial
				material = New PbrMaterial (New Color (Rnd (0.25, 1.0), Rnd (0.25, 1.0), Rnd (0.25, 1.0)), Rnd (1.0), Rnd (1.0))
			Endif
		Endif
		
		model = Model.CreateCone (Cast <RoundVolume> (init.volume).radius, Cast <RoundVolume> (init.volume).length, Axis.Y, segments, material, parent)

		SetColor ()
		
		model.Move (0.0, length * 0.5, 0.0)
		
		ConeList.AddLast (Self)

	End

	Method Start ()

		collider			= New ConeCollider (model)
		body				= New RigidBody (model)
 
		Cast <ConeCollider> (collider).Radius	= Cast <RoundVolume> (init.volume).radius
		Cast <ConeCollider> (collider).Length	= Cast <RoundVolume> (init.volume).length

		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null
		
		live				= True

	End
	
End

Class CylinderBody Extends BodyBase

	Method New (radius:Float = 1.0, length:Float = 1.0, mass:Float = 1.0, material:Material = Null, segments:Int = 16, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= mass
		init.group		= group
		init.mask		= mask

		Local cv:RoundVolume = New RoundVolume
		cv.radius = radius
		cv.length = length
		
		init.volume		= Cast <RoundVolume> (cv)
		
		If Not material
			material = New PbrMaterial (Color.Red)
		Else
			If material = RndMaterial
				material = New PbrMaterial (New Color (Rnd (0.25, 1.0), Rnd (0.25, 1.0), Rnd (0.25, 1.0)), Rnd (1.0), Rnd (1.0))
			Endif
		Endif
		
		model = Model.CreateCylinder (Cast <RoundVolume> (init.volume).radius, Cast <RoundVolume> (init.volume).length, Axis.Y, segments, material, parent)

		SetColor ()

		CylinderList.AddLast (Self)

	End

	Method Start ()

		collider			= New CylinderCollider (model)
		body				= New RigidBody (model)
 
		Cast <CylinderCollider> (collider).Radius	= Cast <RoundVolume> (init.volume).radius
		Cast <CylinderCollider> (collider).Length	= Cast <RoundVolume> (init.volume).length

		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null
		
		live				= True

	End
	
End
Class BoxBody Extends BodyBase

	Method New (width:Float = 1.0, height:Float = 1.0, depth:Float = 1.0, mass:Float = 1.0, material:Material = Null, segments:Int = 1, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= mass
		init.group		= group
		init.mask		= mask
		init.volume		= New Boxf (-width * 0.5, -height * 0.5, -depth * 0.5, width * 0.5, height * 0.5, depth * 0.5)
		
		If Not material
			material = New PbrMaterial (Color.Red)
		Else
			If material = RndMaterial
				material = New PbrMaterial (New Color (Rnd (0.25, 1.0), Rnd (0.25, 1.0), Rnd (0.25, 1.0)), Rnd (1.0), Rnd (1.0))
			Endif
		Endif
		
		model = Model.CreateBox (Cast <Boxf> (init.volume), segments, segments, segments, material, parent)

		SetColor ()

		BoxList.AddLast (Self)

	End

	Method Start ()

		collider			= New BoxCollider (model)
		body				= New RigidBody (model)
 
		Cast <BoxCollider> (collider).Box = Cast <Boxf> (init.volume)
		
		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null
		
		live				= True

	End
	 
End

Class TerrainBody Extends BodyBase

	Field terrain_heightmap:Pixmap
	
	Method New (heightmap:Pixmap, width:Float = 100.0, height:Float = 10.0, depth:Float = 100.0, material:Material = Null, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		If heightmap.Format <> PixelFormat.I8
			Print ""
			Print "*** MEDIA WARNING: Converting heightmap; should be supplied in PixelFormat.I8 format! (See PixelFormat docs.) ***"
			Print ""
			terrain_heightmap = heightmap.Convert (PixelFormat.I8)
		Endif
		
		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= 0.0
		init.group		= group
		init.mask		= mask
		init.volume		= New Boxf (-width * 0.5, -height * 0.5, -depth * 0.5, width * 0.5, height * 0.5, depth * 0.5)
		
		If Not material

			Local ground_pixels:Pixmap = New Pixmap (256, 256, PixelFormat.RGBA8)

				ground_pixels.Clear (Color.Grey)
			
			Local pixel_toggle:Bool = False
			
			For Local gp_y:Int = 0 Until ground_pixels.Height
				For Local gp_x:Int = 0 Until ground_pixels.Width
					If pixel_toggle Then ground_pixels.SetPixel (gp_x, gp_y, Color.White)
					pixel_toggle = Not pixel_toggle
				Next
				pixel_toggle = Not pixel_toggle
			Next
			
			material = New PbrMaterial ()
			Cast <PbrMaterial> (material).ColorTexture = New Texture (ground_pixels, TextureFlags.None)
			
		Else
			If material = RndMaterial
				material = New PbrMaterial (New Color (Rnd (0.25, 1.0), Rnd (0.25, 1.0), Rnd (0.25, 1.0)), Rnd (1.0), Rnd (1.0))
			Endif
		Endif
		
		model = Model.CreateTerrain (terrain_heightmap, Cast <Boxf> (init.volume), material, parent)

		TerrainList.AddLast (Self)

	End

	Method Start ()

		collider			= New TerrainCollider (model)
		body				= New RigidBody (model)

		Cast <TerrainCollider> (collider).Heightmap	= terrain_heightmap
		Cast <TerrainCollider> (collider).Bounds	= Cast <Boxf> (init.volume)
		
		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null
		
		live				= True

	End
	 
End

Class SphereBody Extends BodyBase

	Method New (radius:Float = 0.5, mass:Float = 1.0, material:Material = Null, segments:Int = 32, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= mass
		init.group		= group
		init.mask		= mask
		init.volume		= Cast <Float> (radius)
		
		If Not material
			material = New PbrMaterial (Color.Red)
		Else
			If material = RndMaterial
				material = New PbrMaterial (New Color (Rnd (0.25, 1.0), Rnd (0.25, 1.0), Rnd (0.25, 1.0)), Rnd (1.0), Rnd (1.0))
			Endif
		Endif
		
		model = Model.CreateSphere (radius, segments, segments, material, parent)

		SetColor ()

		SphereList.AddLast (Self)

	End

	Method Start ()

		collider			= New SphereCollider (model)
		body				= New RigidBody (model)
 
		Cast <SphereCollider> (collider).Radius = Cast <Float> (init.volume)
		
		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null

		live				= True
		
	End
	 
End

' Simple flat plane (really a box to allow picking up fast downwards collisions -- per 'height' parameter)...

Class GroundBody Extends BodyBase

	' height value allows for bodies to hit ground at reasonably high speeds
	' without falling through...
	
	Method New (width:Float = 500, height:Float = 10, depth:Float = 500, material:Material = Null, parent:Entity = Null, group:Int = 1, mask:Int = 1)

		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= 0
		init.group		= group
		init.mask		= mask
		init.volume		= New Boxf (-width * 0.5, -height * 0.5, -depth * 0.5, width * 0.5, height * 0.5, depth * 0.5)
		
		If Not material
			material = New PbrMaterial (Color.Green * 0.5)
		Endif
		
		model = Model.CreateBox (Cast <Boxf> (init.volume), 1, 1, 1, material, parent)

		SetColor ()

		' Place top surface at y = 0:
		
		model.Move (0, -height * 0.5, 0)
		
		GroundList.AddLast (Self)

	End

	Method Start ()

		collider			= New BoxCollider (model)
		body				= New RigidBody (model)
 
		Cast <BoxCollider> (collider).Box = Cast <Boxf> (init.volume)
		
		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask

		body.Restitution	= 0.0
		
		init				= Null

		live				= True
		
	End
	
End

' MeshColliders broken -- see Mark's comment in monkey2\modules\mojo3d-vr\tests\projectiles.monkey2!

Global ModelList:List		<MeshBody>		= New List <MeshBody>

Class MeshBody Extends BodyBase

	Method New (pre_made:Model, group:Int = 1, mask:Int = 1)

		'If Not pre_made?.Mesh
		'	RuntimeError ("MeshBody.New requires pre-made model with mesh!")
		'Endif
		
		' Store setup params for body.Start ()
		
		init = New InitParams
		
		init.mass		= 0.0
		init.group		= group
		init.mask		= mask

		model = pre_made

		ModelList.AddLast (Self)

	End

	Method Start ()

		collider			= New MeshCollider (model)
		
		Cast <MeshCollider> (collider).Mesh = model.Mesh
		
		body				= New RigidBody (model)

		body.Mass			= init.mass
		body.CollisionGroup	= init.group
		body.CollisionMask	= init.mask
		body.Restitution	= init.bounce

		init				= Null
		
		live				= True

	End
	
End

Function Shoot (shoot_from:Entity = Null, force:Float = 100.0, size:Float = 2.0, mass:Float = 1.0)
	
	' Create bullet -- mass 5.0 -- as child of
	' camera, to pick up location/rotation...
	
	Local bullet:BoxBody = New BoxBody (size, size, size, mass, New PbrMaterial (Color.Red), 1, shoot_from)

	bullet.SetColor ()
	
	bullet.Move (0.0, 0.0, 2.0)
	
	bullet.model.Parent = Null
	
	bullet.Start ()
	
	bullet.body.ApplyImpulse (bullet.model.Basis * New Vec3f (1.0, 1.0, force))
	
End
	
' Demo

#Rem

	Local car:Model = Model.Load ("asset::car_default.fbx")

		If car
			
			Local carbody:MeshBody = New MeshBody (car)
			
			Local car_scale:Float = 0.05
			
			car.Mesh.FitVertices (New Boxf (-car_scale * (car.Mesh.Bounds.Width * 0.5), -car_scale * (car.Mesh.Bounds.Height * 0.5), -car_scale * (car.Mesh.Bounds.Depth * 0.5), car_scale * (car.Mesh.Bounds.Width * 0.5), car_scale * (car.Mesh.Bounds.Height * 0.5), car_scale * (car.Mesh.Bounds.Depth * 0.5)))
			carbody.Move (0.0, 0.0, 10.0)
			
			carbody.Start ()
			
		Endif
				
#End
